/*****************************************************************************
 ARTIGIANATO ROBOTICO - by Stefano Scibilia
 -----------------------------------------------------------------------------
 Scheda "LAGORINO 1.1" - Primo programma base didattico per Line Following.
 Questa scheda utilizza un arduino nano 3.0 cinese, un modulo bluetooth HC-05, 
 un ponte H, due led, due pulsanti, connessione a barra da 8 sensori IR per 
 line following e dispone di alcuni pin digitali liberi.
 -----------------------------------------------------------------------------
 Nota PWM: i valori dei pwm dei pin 9 e 11 vengono impostati individualmente 
 da pc; i valori dei pin 10 e 3 vengono automaticamente settati complementari
 perchè invertiti in fase di configurazione, dead time non impostato.
 Il setting dei pwm NON fa uso della funzione analogWrite() ma i registri sono
 settati direttamente e non c'è errore di inversione sui valori estremi.
 -----------------------------------------------------------------------------
 Release:
 LF00 - Programma di base minimo:
        
 -----------------------------------------------------------------------------
 Aprire il monitor seriale dalla IDE di arduino, impostando la velocità di 
 trasmissione a 57600 bps; attendere l'avviso di pronto a ricevere e digitare
 i comandi abilitati.
 -----------------------------------------------------------------------------*/

// Assegnazione alias:
#define EnableH 5            // H bridge enable connected to digital pin 5
#define UserButton 8         // left button connected to digital pin 8
#define StartButton 12       // right button connected to digital pin 12
#define RedLed 6             // red led connected to digital pin 6
#define YellowLed 7          // yellow led connected to digital pin 7
#define IR_7_pin A0           // select the input pin for ADC_7
#define IR_6_pin A1           // select the input pin for ADC_6
#define IR_5_pin A2           // select the input pin for ADC_5
#define IR_4_pin A3           // select the input pin for ADC_4
#define IR_3_pin A4           // select the input pin for ADC_3
#define IR_2_pin A5           // select the input pin for ADC_2
#define IR_1_pin A6           // select the input pin for ADC_1
#define IR_0_pin A7           // select the input pin for ADC_0
       
// Dichiarazione variabili globali:
byte PWM_9 = 127;        // memorizza il valore del duty cycle sul pin 9
byte PWM_10 = 127;       // memorizza il valore del duty cycle sul pin 10
byte PWM_11 = 127;       // memorizza il valore del duty cycle sul pin 11
byte PWM_3 = 127;        // memorizza il valore del duty cycle sul pin 3
int K_Proportional = 550;        // proportional 
int K_Derivative = 100;        // derivative
//int K_Integral = 0;           // integral
byte CalibrationSpeed = 180;    // calibration motors speed
byte RunSpeed = 240;   // motors basic speed
byte MaxSpeed = 255;   // motors max speed
int LeftMotorCurrentSpeed = 0;   // left motor current speed
int RightMotorCurrentSpeed = 0;  // right motor current speed
int ValoreRicevuto = 0;       // memorizza il valore di ricevuto sulla seriale
int IR_Vals[8];               // memorizza i valori dei sensori
int IR_Max[8];                // memorizza i valori massimi letti in calibrazione
int IR_Min[8] = {0, 0, 0, 0, 0, 0, 0, 0};  // memorizza i valori minimi letti in calibrazione
int IR_Cal[8] = {100, 100, 100, 100, 100, 100, 100, 100};  // memorizza i fattori di calibrazione
int IR_Vals_Cal[8];           // memorizza i valori calibrati dei sensori
int SensorsCut = 60;          // memorizza il taglio inferiore della lettura dei sensori
byte OnLineFlag = 0;          // flag che mostra se almeno un sensore è sulla linea
byte SensorsShowFlag = 0;     // flag per la trasmissione dei valori dei sensori
byte CalibrationFlag = 0;     // flag per la calibrazione dei sensori
byte RunFlag = 0;             // flag per lo start e lo stop della corsa
signed int LineError = 0;     // valore dell'errore sulla linea
signed int LastLineError = 0; // valore dell'errore precedente
signed int CorrezioneMotori = 0;  // correzione complessiva da gestire con i due motori
int Mass = 0;                 // somma delle letture dei sensori IR
long Torque = 0;              // somma delle letture per il peso della loro posizione

void setup() {
  // put your setup code here, to run once:
  //---------------------------------------------------------
  //-- Configura PWM Timer1 -------------------------------DX
  // Mode 1 - Phase Correct - 8 bit - Inverting mode
   TCCR1A = B10110001;
   // Prescaler = 1
   TCCR1B = B00000001;
   // Imposta il valore del pwm sui pin 9 e 10
   OCR1B = PWM_9;
   OCR1A = PWM_10;
   // Abilita come output i pin 9 e 10
   pinMode(9, OUTPUT);
   pinMode(10, OUTPUT);
  //-- Configura PWM Timer2 -------------------------------SX
  // Mode 1 - Phase Correct - 8 bit - Inverting mode
   TCCR2A = B10110001;
   // Prescaler = 1
   TCCR2B = B00000001;
   // Imposta il valore del pwm sui pin 11 e 3
   OCR2B = PWM_11;
   OCR2A = PWM_3;  
   // Abilita come output i pin 11 e 3
   pinMode(11, OUTPUT);
   pinMode(3, OUTPUT);
  //---------------------------------------------------------
  //-- Pins Configuration --------------------------------------------------
  pinMode(EnableH, OUTPUT);       // sets the digital pin as output
  pinMode(UserButton, INPUT);     // sets the digital pin as input
  pinMode(StartButton, INPUT);    // sets the digital pin as input
  pinMode(RedLed, OUTPUT);        // sets the digital pin as output
  pinMode(YellowLed, OUTPUT);     // sets the digital pin as output  
  //-- Stato iniziale ------------------------------------------------------
  digitalWrite(EnableH, LOW);     // disabilita H bridge    
  CalibrationFlag = 0;            // disabilita la calibrazione dei sensori
  SensorsShowFlag = 0;            // disabilita trasmissione valore sensori IR
  RunFlag = 0;                    // disabilita la corsa
  digitalWrite(YellowLed, LOW);   // tiene spento il led giallo
  digitalWrite(RedLed, LOW);      // tiene spento il led rosso
  //-- Comunicazione iniziale ----------------------------------------------
  Serial.begin(57600);        // opening serial port
  delay(1000);                // wait for PC
  Serial.println(F(" ARTIGIANATO ROBOTICO by Stefano Scibilia "));
  Serial.println(F("------------------------------------------"));
  Serial.println(F(" Setting Lagorino per Line Following      "));
  Serial.println(F("------------------------------------------"));
  Serial.println(F("Digitare: p<int>      (set K proporzional)"));
  Serial.println(F("          d<int>        (set K derivative)"));
  //Serial.println(F("          i<int>          (set K integral)"));
  Serial.println(F("          t<byte>  (set calibration speed)"));
  Serial.println(F("          l<byte>          (set run speed)"));
  Serial.println(F("          L<byte>          (set max speed)")); 
  Serial.println(F("          k<byte>            (sensors cut)"));
  Serial.println(F("          c          (sensors calibration)"));
  Serial.println(F("          s          (enable start button)"));
  Serial.println(F("          S            (remote start/stop)"));
  Serial.println(F("          E              (enable H bridge)"));
  Serial.println(F("          e             (disable H bridge)"));
  Serial.println(F("          X<byte>   (set speed left motor)"));
  Serial.println(F("          Y<byte>  (set speed right motor)"));
  Serial.println(F("          Z        (show sensors value ON)"));
  Serial.println(F("          z       (show sensors value OFF)"));
  Serial.println(F("          ?       (mostra valori correnti)")); 
  Serial.println(F("------------------------------------------"));    
  //------------------------------------------------------------------------
}

void loop() {
  // put your main code here, to run repeatedly:
  //controllo se c'è qualcosa in arrivo sulla seriale
  if (Serial.available()) {  
    GestisciRicezioneSeriale();
    }
  // --- Sensori ed errore -----------------------------------------------------  
  LeggiSensoriIRxLF();
  NormalizzaSensoriIR();
  CalcolaErrore();
  if (RunFlag == 1) {
    RunLoop();
    }    
  if (CalibrationFlag == 1) {
    CalibrationLoop();
  }
  // ---------------------------------------------------------------------------   
}

// put your function here:
void GestisciRicezioneSeriale() {
  byte command = Serial.read(); //leggo il primo byte
  switch (command)  //controllo che sia un comando valido
    {
    case '?': // è richiesto lo stato delle periferiche
      Serial.println(F("------------------------------------------"));
      Serial.print(F("K proportional impostato a   "));
      Serial.println(K_Proportional, DEC);       
      Serial.print(F("K derivative impostato a     "));
      Serial.println(K_Derivative, DEC);      
      Serial.print(F("CalibrationSpeed impostata a "));
      Serial.println(CalibrationSpeed, DEC);  
      Serial.print(F("RunSpeed impostata a         "));
      Serial.println(RunSpeed, DEC);        
      Serial.print(F("MaxSpeed impostata a         "));
      Serial.println(MaxSpeed, DEC);      
      Serial.print(F("Sensors cut impostato a      "));
      Serial.println(SensorsCut, DEC); 
      Serial.println(F("------------------------------------------"));
      break;
    case 'p': // trasmissione nuovo valore di K_Proporzional
      RaccoltaValoreRicevuto();
      // se il valore di K_Proporzional è valido, lo memorizzo nella variabile
      if (ValoreRicevuto <= 1000) {
        K_Proportional = ValoreRicevuto;
        Serial.print(F("K proportional impostato a "));
        Serial.println(K_Proportional, DEC);
        }
      break;  
    case 'd': // trasmissione nuovo valore di K_Derivative
      RaccoltaValoreRicevuto();
      // se il valore di K_Derivative è valido, lo memorizzo nella variabile
      if (ValoreRicevuto <= 10000) {
        K_Derivative = ValoreRicevuto;
        Serial.print(F("K derivative impostato a "));
        Serial.println(K_Derivative, DEC);
        }
      break;  
//      case 'i': // trasmissione nuovo valore di K_Integral
//        RaccoltaValoreRicevuto();
//        // se il valore di K_Integral è valido, lo memorizzo nella variabile
//        if (ValoreRicevuto <= 10000) {
//          K_Integral = ValoreRicevuto;
//          Serial.print(F("K integral impostato a "));
//          Serial.println(K_Integral, DEC);
//          }
//        break;   
    case 't': // trasmissione valore CalibrationSpeed
      RaccoltaValoreRicevuto();
      // se il valore di CalibrationSpeed è valido, lo memorizzo nella variabile
      if (ValoreRicevuto <= 255) {
        CalibrationSpeed = ValoreRicevuto;
        Serial.print(F("CalibrationSpeed impostata a "));
        Serial.println(CalibrationSpeed, DEC);  
        }
      break;   
    case 'l': // trasmissione valore RunSpeed
      RaccoltaValoreRicevuto();
      // se il valore di RunSpeed è valido, lo memorizzo nella variabile
      if (ValoreRicevuto <= 255) {
        RunSpeed = ValoreRicevuto;
        Serial.print(F("RunSpeed impostata a "));
        Serial.println(RunSpeed, DEC);  
        }
      break;  
    case 'L': // trasmissione valore MaxSpeed
      RaccoltaValoreRicevuto();
      // se il valore di MaxSpeed è valido, lo memorizzo nella variabile
      if (ValoreRicevuto <= 255) {
        MaxSpeed = ValoreRicevuto;
        Serial.print(F("MaxSpeed impostata a "));
        Serial.println(MaxSpeed, DEC);
        }
      break;  
    case 'k': // trasmissione valore SensorsCut
      RaccoltaValoreRicevuto();
      // se il valore di SensorsCut è valido, lo memorizzo nella variabile
      if (ValoreRicevuto <= 255) {
        SensorsCut = ValoreRicevuto;
        Serial.print(F("SensorsCut impostato a "));
        Serial.println(SensorsCut, DEC);
        }
      break;    
    case 'c': // enable calibration loop
      CalibrationFlag = 1;
      break;   
    case 'S': // enable/disable RUN
      if(RunFlag == 0) {
        RunFlag = 1;
        ControllaMotoreSx(127);
        ControllaMotoreDx(127);
        digitalWrite(EnableH, HIGH);     // abilita H bridge      
        digitalWrite(RedLed, HIGH);      // tiene acceso il led rosso
        delay(200);
        }
      else {
        RunFlag = 0;
        // ferma il robot
        ControllaMotoreSx(127);
        ControllaMotoreDx(127); 
        delay(1000);
        digitalWrite(EnableH, LOW);     // disabilita H bridge      
        digitalWrite(RedLed, LOW);      // tiene spento il led rosso
        }         
      break;           
    case 'E': // enable H bridge
      digitalWrite(EnableH, HIGH);
      Serial.println(F("H bridge abilitato"));
      break;    
    case 'e': // disable H bridge
      digitalWrite(EnableH, LOW);
      Serial.println(F("H bridge disabilitato"));
      break;        
    case 'X': // trasmissione valore velocità motore sinistro
      RaccoltaValoreRicevuto();
      // se il valore è valido, imposto la velocità
      if (ValoreRicevuto <= 255) {
        ControllaMotoreSx(ValoreRicevuto);
        Serial.print(F("PWM pin 11 impostato a "));
        Serial.println(PWM_11, DEC);
        Serial.print(F("PWM pin  3 impostato a "));
        Serial.println(PWM_3, DEC);
        }        
      break;    
    case 'Y': // trasmissione valore velocità motore destro
      RaccoltaValoreRicevuto();
      // se il valore è valido, imposto la velocità
      if (ValoreRicevuto <= 255) {
        ControllaMotoreDx(ValoreRicevuto);
        Serial.print(F("PWM pin  9 impostato a "));
        Serial.println(PWM_9, DEC);
        Serial.print(F("PWM pin 10 impostato a "));
        Serial.println(PWM_10, DEC);
        }
      break;   
    case 'Z': // enable sensors value transmission
      SensorsShowFlag = 1;
      break;
    case 'z': // disable sensors value transmission 
      SensorsShowFlag = 0;
      break;           
    //------------------------------------------------------------------------
   }
  SvuotaBuffer();
}

void RaccoltaValoreRicevuto() {
 unsigned long Millisecondi;  // serve per il timeout sulla ricezione del valore
 byte QuantiCaratteri;        // memorizza il numero di caratteri ricevuti
 byte Carattere;              // memorizza il carattere letto 
 ValoreRicevuto = 0;
 Millisecondi = millis();
 QuantiCaratteri = 0;
 // loop per raccogliere i caratteri del valore
 do {
  if (Serial.available()) {
    Carattere = Serial.read();
    QuantiCaratteri++;
    if ((Carattere >= 48) && (Carattere <= 57)) {      // se il carattere rappresenta un numero        
      ValoreRicevuto = (ValoreRicevuto * 10) + (Carattere - 48);
      } 
    else if ((Carattere == 10) || (Carattere == 13)) { //se il carattere rappresenta [invio] e/o [a capo]      
      break;
      }
    }
  } 
 // esco per timeout, per un valore  maggiore di 255, oppure per troppi caratteri senza senso ricevuti
 while((millis() - Millisecondi < 500) && (ValoreRicevuto <= 32767) && (QuantiCaratteri < 10));
 }

void SvuotaBuffer() {
 //svuoto il buffer da eventuali altri caratteri che non mi servono più
 while (Serial.available()) {
   byte a = Serial.read();
   }
 }

void ControllaMotoreDx(byte Val) { // controlla direzione e velocità motore sinistro     
  PWM_9 = Val;
  PWM_10 = Val;  
  OCR1A = PWM_10;
  OCR1B = PWM_9;  
  }

void ControllaMotoreSx(byte Val) { // controlla direzione e velocità motore destro    
  PWM_11 = Val;
  PWM_3 = Val;  
  OCR2A = PWM_3;
  OCR2B = PWM_11;  
  }  

void LeggiSensoriIRxLF() {
  OnLineFlag = 0;   // azzero il flag che segnala se almeno un sensore è sulla linea
  IR_Vals[0] = analogRead(IR_0_pin);
  IR_Vals[1] = analogRead(IR_1_pin);
  IR_Vals[2] = analogRead(IR_2_pin);
  IR_Vals[3] = analogRead(IR_3_pin);
  IR_Vals[4] = analogRead(IR_4_pin);
  IR_Vals[5] = analogRead(IR_5_pin);
  IR_Vals[6] = analogRead(IR_6_pin);
  IR_Vals[7] = analogRead(IR_7_pin);
  for (int i = 0; i < 8; i++) {
    if (IR_Vals[i] > 300) OnLineFlag = 1;
  }
}

void NormalizzaSensoriIR() {
  for (int i = 0; i < 8; i++) {
    IR_Vals_Cal[i] = ((long)IR_Vals[i] - IR_Min[i]) * IR_Cal[i]/100;      
    }
}

void CalcolaErrore() {
  if (OnLineFlag == 1) {
    Mass = 0;
    Torque = 0;
    for (int i = 0; i < 8; i++) {
      if(IR_Vals_Cal[i] > SensorsCut) {
        Mass = Mass + IR_Vals_Cal[i];
        Torque = Torque + ((long)IR_Vals_Cal[i]*1000*i);
      }
    }  
  }    
  LineError = 3500 - (Torque/Mass);   
  if (SensorsShowFlag == 1) {
    for (byte i = 0; i < 8; i++) {
     Serial.print(IR_Vals_Cal[i], DEC);
     Serial.print(F("."));
    }
    Serial.print(Mass, DEC);
    Serial.print(F("."));
    Serial.print(Torque, DEC);
    Serial.print(F("."));
    Serial.println(LineError, DEC);
  } 
}

void TrovaMinMax() {
  for (int i = 0; i < 8; i++) {
    IR_Max[i] = max(IR_Max[i], IR_Vals[i]);   
    IR_Min[i] = min(IR_Min[i], IR_Vals[i]);    
    } 
}

void CalcolaFattoreDiCalibrazione() {
  for (int i = 0; i < 8; i++) {
    IR_Cal[i] = 100000/(IR_Max[i]- IR_Min[i]);   
    Serial.println(IR_Cal[i], DEC);  
    } 
}

void CalibrationLoop() {
  digitalWrite(YellowLed, HIGH);
  digitalWrite(EnableH, HIGH);
  for (int i = 0; i < 8; i++) {
    IR_Max[i] = 0;                           // preimposta valore minimo
    IR_Min[i] = 1000;                        // preimposta valore massimo  
    }
  delay(200);                // wait for 200ms
  while (OnLineFlag == 0) { 
    LeggiSensoriIRxLF();    // leggi i sensori
    } // attendi che il robot sia sulla linea 
  // ruota robot a sinistra  
  ControllaMotoreSx(255 - CalibrationSpeed);
  ControllaMotoreDx(CalibrationSpeed);
  while (OnLineFlag == 1) { // fino a che il robot è sulla linea ruota a sinistra
    LeggiSensoriIRxLF();    // leggi i sensori
    TrovaMinMax();          // individua le letture massima e minima per ogni sensore
    }
  while (OnLineFlag == 0) { // fino a che il robot è fuori linea ruota a sinistra
    LeggiSensoriIRxLF();    // leggi i sensori
    TrovaMinMax();          // individua le letture massima e minima per ogni sensore
    }
  delay(200);                // wait for 200ms  
  // ruota robot a destra  
  ControllaMotoreSx(CalibrationSpeed);
  ControllaMotoreDx(255 - CalibrationSpeed);
  while (OnLineFlag == 1) { // fino a che il robot è sulla linea ruota a destra
    LeggiSensoriIRxLF();    // leggi i sensori
    TrovaMinMax();          // individua le letture massima e minima per ogni sensore
    }
  while (OnLineFlag == 0) { // fino a che il robot è fuori linea ruota a destra
    LeggiSensoriIRxLF();    // leggi i sensori
    TrovaMinMax();          // individua le letture massima e minima per ogni sensore
    }
  while (OnLineFlag == 1) { // fino a che il robot è sulla linea ruota a destra
    LeggiSensoriIRxLF();    // leggi i sensori
    TrovaMinMax();          // individua le letture massima e minima per ogni sensore
    }  
  while (OnLineFlag == 0) { // fino a che il robot è fuori linea ruota a destra
    LeggiSensoriIRxLF();    // leggi i sensori
    TrovaMinMax();          // individua le letture massima e minima per ogni sensore
    }
  delay(200);                // wait for 200ms    
  // ruota robot a sinistra  
  ControllaMotoreSx(255 - CalibrationSpeed);
  ControllaMotoreDx(CalibrationSpeed);
  while (OnLineFlag == 1) { // fino a che il robot è sulla linea ruota a sinistra
    LeggiSensoriIRxLF();    // leggi i sensori
    TrovaMinMax();          // individua le letture massima e minima per ogni sensore
    }
  while (OnLineFlag == 0) { // fino a che il robot è fuori linea ruota a sinistra
    LeggiSensoriIRxLF();    // leggi i sensori
    TrovaMinMax();          // individua le letture massima e minima per ogni sensore
    } 
  CalcolaFattoreDiCalibrazione();   
  do {  // fino a che l'errore è alto
    LeggiSensoriIRxLF();    // leggi i sensori
    NormalizzaSensoriIR();
    CalcolaErrore();
    if(LineError > 0) {
      ControllaMotoreSx(255 - CalibrationSpeed);
      ControllaMotoreDx(CalibrationSpeed);
      }
    else if(LineError < 0) {
      ControllaMotoreSx(CalibrationSpeed);
      ControllaMotoreDx(255 - CalibrationSpeed);
      } 
    }while (LineError > 10 || LineError < -10);    
  // ferma il robot
  ControllaMotoreSx(127);
  ControllaMotoreDx(127); 
  delay(200);                // wait for 200ms   
  CalibrationFlag = 0;
  digitalWrite(EnableH, LOW);
  digitalWrite(YellowLed, LOW);
}

void RunLoop() {
  CorrezioneMotori = (long)K_Proportional * LineError / 10000 + (long)K_Derivative * (LineError - LastLineError) / 100;
  LastLineError = LineError;  // tieni memoria dell'errore precedente
  LeftMotorCurrentSpeed = RunSpeed - CorrezioneMotori;
  RightMotorCurrentSpeed = RunSpeed + CorrezioneMotori;  
  if (LeftMotorCurrentSpeed < 0) {
    LeftMotorCurrentSpeed = 0;     // impedisce che possa essere un valore negativo      
  }
  if (RightMotorCurrentSpeed < 0) {
    RightMotorCurrentSpeed = 0;     // impedisce che possa essere un valore negativo      
  }
  if (LeftMotorCurrentSpeed > MaxSpeed) {
    LeftMotorCurrentSpeed = MaxSpeed;     // impedisce che possa essere un valore superiore alla massima velocità impostata        
  }
  if (RightMotorCurrentSpeed > MaxSpeed) {
    RightMotorCurrentSpeed = MaxSpeed;     // impedisce che possa essere un valore superiore alla massima velocità impostata      
  }
  ControllaMotoreSx(LeftMotorCurrentSpeed);
  ControllaMotoreDx(RightMotorCurrentSpeed); 
}

