/*****************************************************************************
 ARTIGIANATO ROBOTICO - by Stefano Scibilia
 -----------------------------------------------------------------------------
 Programma di test della scheda di controllo del robot didattico "Lagorino".
 Questa scheda utilizza un arduino nano 3.0 cinese, un modulo xBee, ponte H,
 due led, due pulsanti, connessione a barra da 8 sensori IR per line following
 e alcuni pin digitali liberi.
 Implementati segnali pwm su Timer1 (pin 9 e 10) e Timer2 (pin 11 e 3).
 - p9<byte> assegna il valore pwm sul pin 9
 - p10<byte> assegna il valore pwm sul pin 10
 - p11<byte> assegna il valore pwm sul pin 11
 - p3<byte> assegna il valore pwm sul pin 3
 -----------------------------------------------------------------------------
 Nota PWM: i valori dei pwm dei pin 9 e 11 vengono impostati individualmente 
 da pc; i valori dei pin 10 e 3 vengono automaticamente settati complementari
 perchè invertiti in fase di configurazione, dead time non impostato.
 Il setting dei pwm NON fa uso della funzione analogWrite() ma i registri sono
 settati direttamente e non c'è errore di inversione sui valori estremi.
 -----------------------------------------------------------------------------
 Aprire il monitor seriale dalla IDE di arduino, impostando la velocità di 
 trasmissione a 9600 (la prima voltaprima della configurazione dell'HC-05) o
 57600 bps (dopo configurazione); attendere l'avviso di pronto a ricevere e 
 digitare i comandi abilitati.
 */

 // Assegnazione alias
 // const byte DeadTime = 2;   // imposta il valore di DeadTime tra i PWM complementari
 
 // Dichiarazione variabili globali
 byte PWM_9 = 127;        // memorizza il valore del duty cycle sul pin 9
 byte PWM_10 = 127;       // memorizza il valore del duty cycle sul pin 10
 byte PWM_11 = 127;       // memorizza il valore del duty cycle sul pin 11
 byte PWM_3 = 127;        // memorizza il valore del duty cycle sul pin 3
 int ValoreRicevuto = 0;  // memorizza il valore di PWM letto dalla seriale
 int LedRosso = 6;        // red led connected to digital pin 6
 int LedGiallo = 7;       // yellow led connected to digital pin 7
 int EnableH = 5;         // H bridge enable connected to digital pin 5
 int PulsanteSX = 8;      // left button connected to digital pin 8
 int PulsanteDX = 12;     // right button connected to digital pin 12
 int StatoPulsanteSX;     // left button state
 int StatoPulsanteDX;     // right button state
 int ADC_Flag = 0;        // segnala se la trasmissione dei valori è attiva
 int ADC_0_Pin = A0;      // select the input pin for  ADC_0
 int ADC_1_Pin = A1;      // select the input pin for  ADC_1
 int ADC_2_Pin = A2;      // select the input pin for  ADC_2
 int ADC_3_Pin = A3;      // select the input pin for  ADC_3
 int ADC_4_Pin = A4;      // select the input pin for  ADC_4
 int ADC_5_Pin = A5;      // select the input pin for  ADC_5
 int ADC_6_Pin = A6;      // select the input pin for  ADC_6
 int ADC_7_Pin = A7;      // select the input pin for  ADC_7
 int ADC_0;               // memorizza il valore letto su a0
 int ADC_1;               // memorizza il valore letto su a1
 int ADC_2;               // memorizza il valore letto su a2
 int ADC_3;               // memorizza il valore letto su a3
 int ADC_4;               // memorizza il valore letto su a4
 int ADC_5;               // memorizza il valore letto su a5
 int ADC_6;               // memorizza il valore letto su a6
 int ADC_7;               // memorizza il valore letto su a7
 
 void setup()
   {
 //---------------------------------------------------------
 //-- Configura PWM Timer1 ---------------------------------
 // Mode 1 - Phase Correct - 8 bit - Inverting mode
   TCCR1A = B10110001;
   // Prescaler = 1
   TCCR1B = B00000001;
   // Imposta il valore del pwm sui pin 9 e 10
   OCR1B = PWM_9;
   OCR1A = PWM_10;
   // Abilita come output i pin 9 e 10
   pinMode(9, OUTPUT);
   pinMode(10, OUTPUT);
 //-- Configura PWM Timer2 ---------------------------------
 // Mode 1 - Phase Correct - 8 bit - Inverting mode
   TCCR2A = B10110001;
   // Prescaler = 1
   TCCR2B = B00000001;
   // Imposta il valore del pwm sui pin 11 e 3
   OCR2B = PWM_11;
   OCR2A = PWM_3;  
   // Abilita come output i pin 11 e 3
   pinMode(11, OUTPUT);
   pinMode(3, OUTPUT);
 //---------------------------------------------------------
 //-- Configura Pins ---------------------------------------
   pinMode(LedRosso, OUTPUT);      // sets the digital pin as output
   pinMode(LedGiallo, OUTPUT);     // sets the digital pin as output
   pinMode(EnableH, OUTPUT);       // sets the digital pin as output
   pinMode(PulsanteSX, INPUT);     // sets the digital pin as input
   pinMode(PulsanteDX, INPUT);     // sets the digital pin as input
  
   Serial.begin(57600); // apro la seriale
//   Serial.begin(9600); // apro la seriale
   delay(1000); //attendo che l'utente faccia altrettanto
   Serial.println(F(" ARTIGIANATO ROBOTICO by Stefano Scibilia ")); // sono pronto
   Serial.println(F("------------------------------------------"));
   Serial.println(F("Test PWM su arduino Nano connesso e pronto")); 
   Serial.println(F("------------------------------------------"));
   Serial.println(F("Digitare: e<bit>   (disable<0>  enable<1>)"));
   Serial.println(F("          a<byte>  (set duty cycle pin  9)"));
   Serial.println(F("                   (set duty cycle pin 10)")); 
   Serial.println(F("          c<byte>  (set duty cycle pin 11)")); 
   Serial.println(F("                   (set duty cycle pin  3)")); 
   Serial.println(F("          g<bit> (led giallo off<0> on<1>)")); 
   Serial.println(F("          r<bit> (led rosso  off<0> on<1>)")); 
   Serial.println(F("          s<bit> (visualizza valori 8 adc)")); 
   Serial.println(F("          ?       (mostra valori correnti)")); 
   Serial.println(F("------------------------------------------"));
  }

 //programma principale
 void loop() 
  {
  //controllo se c'è qualcosa in arrivo sulla seriale
  if (Serial.available()) 
    {
    byte command = Serial.read(); //leggo il primo byte
    switch (command)  //controllo che sia un comando valido
      {
      case '?': // è richiesto lo stato delle periferiche
        Serial.println(F("------------------------------------------"));
        Serial.print(F("PWM pin  9 impostato a "));
        Serial.print(PWM_9, DEC);       
        Serial.println(F(" su 255;"));
        Serial.print(F("PWM pin 10 impostato a "));
        Serial.print(PWM_10, DEC);       
        Serial.println(F(" su 255;"));
        Serial.print(F("PWM pin 11 impostato a "));
        Serial.print(PWM_11, DEC);       
        Serial.println(F(" su 255;"));
        Serial.print(F("PWM pin  3 impostato a "));
        Serial.print(PWM_3, DEC);       
        Serial.println(F(" su 255;"));      
        Serial.println(F("------------------------------------------"));
        break;
      case 'a': // trasmissione valore PWM da impostare sul pin 9
        RaccoltaValoreRicevuto();
        // se il valore di PWM è valido, lo memorizzo nella variabile relativa e lo imposto sulla periferica
        if (ValoreRicevuto <= 255) 
          {
          PWM_9 = ValoreRicevuto;
          PWM_10 = ValoreRicevuto;
          OCR1A = PWM_10;
          OCR1B = PWM_9;
          Serial.print(F("PWM pin  9 impostato a "));
          Serial.println(PWM_9, DEC);
          Serial.print(F("PWM pin 10 impostato a "));
          Serial.println(PWM_10, DEC);
          }
        break;  
      case 'c': // trasmissione valore PWM da impostare sul pin 11
        RaccoltaValoreRicevuto();
        // se il valore di PWM è valido, lo memorizzo nella variabile relativa e lo imposto sulla periferica
        if (ValoreRicevuto <= 255) 
          {
          PWM_11 = ValoreRicevuto;
          PWM_3 = ValoreRicevuto;
          OCR2A = PWM_3;
          OCR2B = PWM_11;
          Serial.print(F("PWM pin 11 impostato a "));
          Serial.println(PWM_11, DEC);
          Serial.print(F("PWM pin  3 impostato a "));
          Serial.println(PWM_3, DEC);
          }
        break;    
      case 'e': // trasmissione comando enable/disable motori
        RaccoltaValoreRicevuto();
        // se il valore ricevuto è valido, eseguo l'abilitazione o la disabilitazione dei motori
        if (ValoreRicevuto == 1) 
          {
          digitalWrite(EnableH, HIGH);   // abilita i motori
          }
        if (ValoreRicevuto == 0) 
          {
          digitalWrite(EnableH, LOW);   // disabilita i motori
          }
        break;           
      case 'r': // trasmissione comando acceso/spento led rosso
        RaccoltaValoreRicevuto();
        // se il valore ricevuto è valido, eseguo l'accensione o lo spegnimento del led rosso
        if (ValoreRicevuto == 1) 
          {
          digitalWrite(LedRosso, HIGH);   // accende il led rosso
          }
        if (ValoreRicevuto == 0) 
          {
          digitalWrite(LedRosso, LOW);   // spegne il led rosso
          }
        break;       
     case 's': // trasmissione comando attivato/disattivato l'invio delle letture analogiche
        RaccoltaValoreRicevuto();
        if (ValoreRicevuto == 1) 
          {
          ADC_Flag = 1;  
          }
        if (ValoreRicevuto == 0) 
          {
          ADC_Flag = 0;   
          }
        break;            
     case 'g': // trasmissione comando acceso/spento led giallo
        RaccoltaValoreRicevuto();
        // se il valore ricevuto è valido, eseguo l'accensione o lo spegnimento del led giallo
        if (ValoreRicevuto == 1) 
          {
          digitalWrite(LedGiallo, HIGH);   // accende il led giallo
          }
        if (ValoreRicevuto == 0) 
          {
          digitalWrite(LedGiallo, LOW);   // spegne il led giallo
          }
        break;                  
     }
   SvuotaBuffer();
   }
   
// StatoPulsanteSX = digitalRead(PulsanteSX);   // read the input pin  
// StatoPulsanteDX = digitalRead(PulsanteDX);   // read the input pin  
// digitalWrite(LedGiallo, StatoPulsanteSX);    // accende il led giallo se è premuto il pulsante SX, altrimenti lo spegne
// digitalWrite(LedRosso, StatoPulsanteDX);     // accende il led rosso se è premuto il pulsante DX, altrimenti lo spegne

// read value from sensors:
  ADC_0 = analogRead(ADC_0_Pin);
  ADC_1 = analogRead(ADC_1_Pin);
  ADC_2 = analogRead(ADC_2_Pin);
  ADC_3 = analogRead(ADC_3_Pin);
  ADC_4 = analogRead(ADC_4_Pin);
  ADC_5 = analogRead(ADC_5_Pin);
  ADC_6 = analogRead(ADC_6_Pin);
  ADC_7 = analogRead(ADC_7_Pin);
  if (ADC_Flag == 1) 
    {
    Serial.print(ADC_0, DEC);
    Serial.print(F(" - "));      
    Serial.print(ADC_1, DEC);
    Serial.print(F(" - ")); 
    Serial.print(ADC_2, DEC); 
    Serial.print(F(" - ")); 
    Serial.print(ADC_3, DEC);
    Serial.print(F(" - ")); 
    Serial.print(ADC_4, DEC); 
    Serial.print(F(" - "));    
    Serial.print(ADC_5, DEC);
    Serial.print(F(" - ")); 
    Serial.print(ADC_6, DEC); 
    Serial.print(F(" - "));  
    Serial.println(ADC_7, DEC);
    }
 }

 void RaccoltaValoreRicevuto()
 {
 unsigned long Millisecondi;  // serve per il timeout sulla ricezione del valore di PWM
 byte QuantiCaratteri;        // memorizza il numero di caratteri ricevuti
 byte Carattere;              // memorizza il carattere letto 
 ValoreRicevuto = 0;
 Millisecondi = millis();
 QuantiCaratteri = 0;
 // loop per raccogliere i caratteri del valore di PWM
 do 
  {
  if (Serial.available())
    {
    Carattere = Serial.read();
    QuantiCaratteri++;
    if ((Carattere >= 48) && (Carattere <= 57))       // se il carattere rappresenta un numero 
      { 
      ValoreRicevuto = (ValoreRicevuto * 10) + (Carattere - 48);
      } 
    else if ((Carattere == 10) || (Carattere == 13))  //se il carattere rappresenta [invio] e/o [a capo]
      {
      break;
      }
    }
  } 
 // esco per timeout, per un valore  maggiore di 255, oppure per troppi caratteri senza senso ricevuti
 while((millis() - Millisecondi < 500) && (ValoreRicevuto <= 255) && (QuantiCaratteri < 10));
 }

 void SvuotaBuffer()
 {
 //svuoto il buffer da eventuali altri caratteri che non mi servono più
 while (Serial.available())
   {
    byte a = Serial.read();
   }
 }

